module GraphicsWindow
	( module GraphicsWindow
	) where

import qualified Xlib as X
import Bits
import Word
import IOExts
import Monad( unless, when )
import Concurrent
	( forkIO
	)
import System( getEnv )
import IO( try )
import Maybe
import Char( intToDigit )
import Word( word32ToInt )
import qualified Graphics_Utilities as Utils
import GraphicsEvent
import qualified GraphicsEvents as E
import GraphicsDC
import qualified GraphicsTimer as T

----------------------------------------------------------------
-- Interface
----------------------------------------------------------------

type Title = String

runGraphics :: IO () -> IO ()  -- SOE, p48

data Window = MkWindow  
  { wnd     :: X.Window	         -- the real window
  , ref_dc  :: IORef (Maybe DC)  -- "device context"
  , events  :: E.Events	         -- the event stream
  , graphic :: IORef Graphic     -- the current graphic
  , redraw  :: RedrawStuff
  , timer   :: Maybe T.Timer
  }
  
openWindowEx :: Title -> Maybe Point -> Maybe Size -> 
                RedrawMode -> Maybe T.Time -> IO Window

closeWindow    :: Window -> IO ()
getWindowRect  :: Window -> IO (Point,Point)
getWindowEvent :: Window -> IO Event
maybeGetWindowEvent :: Window -> IO (Maybe Event)

type Graphic = Draw ()

-- note that modGraphic doesn't force a redraw
setGraphic :: Window -> Graphic -> IO ()
getGraphic :: Window -> IO Graphic
modGraphic :: Window -> (Graphic -> Graphic) -> IO ()
directDraw :: Window -> Graphic -> IO ()

getTime    :: IO T.Time

----------------------------------------------------------------
-- Implementation
----------------------------------------------------------------

runGraphics m = do
  disp <- try (getEnv "DISPLAY")
  runGraphicsEx (either (const ":0.0") id disp) m

----------------------------------------------------------------
-- Windows
----------------------------------------------------------------

closeWindow' :: Bool -> Window -> IO ()
closeWindow' destroyXWindow w = do
  mb_dc <- readIORef (ref_dc w)
  case mb_dc of
   Just dc -> do
     writeIORef (ref_dc w) Nothing -- mark it for dead
     removeDeadWindows             -- bring out your dead
     X.freeGC (disp dc) (textGC  dc)
     X.freeGC (disp dc) (paintGC dc)
     X.freeGC (disp dc) (brushGC dc)
     when destroyXWindow $ do
       X.destroyWindow (disp dc) (drawable dc)  -- ths dc had better hold a window!
       minor_eloop (disp dc)
   Nothing -> do
     return ()

removeDeadWindows :: IO ()
removeDeadWindows = do
  ws <- readIORef wnds
  ws' <- remove ws []
  writeIORef wnds ws'
 where
  remove [] r = return r
  remove (w:ws) r = do
    mb_dc <- readIORef (ref_dc w)
    if (isJust mb_dc) 
      then remove ws (w:r)
      else remove ws r

closeAllWindows :: IO ()
closeAllWindows = do
  ws <- readIORef wnds
  mapM_ closeWindow ws

sendTicks :: IO ()
sendTicks = do
  ws <- readIORef wnds
  sequence_ [ E.sendTick (events w) | w <- ws ]

getWindowTick :: Window -> IO ()
getWindowTick w = E.getTick (events w)

-- persistent list of open windows
wnds :: IORef [Window]
wnds = unsafePerformIO (newIORef [])

-- persistent list of timers
timers :: T.Timers
timers = unsafePerformIO T.newTimers

runGraphicsEx :: String -> IO () -> IO ()
runGraphicsEx host m = do
  X.setDefaultErrorHandler
  display <- X.openDisplay host `catch` \ err -> 
            ioError (userError ("Unable to open X display " ++ host))
  writeIORef displayRef display
  T.clearTimers timers
--  color_map <- X.getStandardColormap display root X.a_RGB_BEST_MAP
  forkIO m
  main_eloop display
--   X.sync display True
--   closeAllWindows
--   X.sync display True
  X.closeDisplay display
  writeIORef displayRef (error "Display not opened yet")

showHex :: Integral a => a -> ShowS
showHex x r = if h == 0 then r' else showHex h r'
 where
  (h,l) = x `divMod` 16
  r'    = intToDigit (fromIntegral l) : r

-- This is what we use on Win32
-- type RedrawMode = Graphic -> DrawFun
-- drawGraphic            ::                    RedrawMode
-- drawBufferedGraphic    ::                    RedrawMode
-- drawBufferedGraphicBC  :: Win32.COLORREF  -> RedrawMode

-- On X, we're a bit more abstract
data RedrawMode
  = Unbuffered
  | DoubleBuffered

----------------------------------------------------------------
-- Implementation
----------------------------------------------------------------

openWindowEx name pos size redrawMode tickRate = do
  display <- readIORef displayRef
  let
    (x,y) = fromPoint (fromMaybe (0,0) pos)
    (w,h) = fromSize  (fromMaybe (0,0) size) 
  let screen    = X.defaultScreenOfDisplay display
      fg_color  = X.whitePixelOfScreen screen
      bg_color  = X.blackPixelOfScreen screen
      depth     = X.defaultDepthOfScreen screen
      root      = X.rootWindowOfScreen screen
      visual    = X.defaultVisualOfScreen screen

  -- ToDo: resurrect the old code for constructing attribute sets
  attributes <- X.allocXSetWindowAttributes
  X.set_background_pixel attributes bg_color  
  let event_mask = 
     	(   X.buttonPressMask
     	.|. X.buttonReleaseMask
     	.|. X.keyPressMask
     	.|. X.keyReleaseMask
     	.|. X.pointerMotionMask
     	.|. X.exposureMask
     	.|. X.structureNotifyMask
     	)
  X.set_event_mask attributes event_mask
  -- We use backing store to reduce the number of expose events due to
  -- raising/lowering windows.
  X.set_backing_store attributes X.whenMapped
  -- We use bit-gravity to avoid generating exposure events when a window is
  -- made smaller (they can't be avoided when the window is enlarged).
  -- The choice of NW is somewhat arbitrary but hopefully works often
  -- enough to be worth it.
  X.set_bit_gravity attributes X.northWestGravity
  let attrmask 
        =   X.cWBackPixel 
        .|. X.cWEventMask 
        .|. X.cWBackingStore
        .|. X.cWBitGravity
  window <- X.createWindow display root 
	 x y -- x, y
	 w h -- width, height
	 1   -- border_width
	 depth          -- use CopyFromParent??
	 X.inputOutput
	 visual         -- use CopyFromParent??
	 attrmask
	 attributes
  X.free attributes

  -- AC, 1/9/2000: Tell the window manager that we want to use the
  -- DELETE_WINDOW protocol
  delWinAtom <- X.internAtom display "WM_DELETE_WINDOW" False
  X.setWMProtocols display window [delWinAtom]
 
  X.setTextProperty display window name	X.a_WM_ICON_NAME 
  X.setTextProperty display window name	X.a_WM_NAME 
  X.mapWindow display window 
  X.raiseWindow display window

  text_gc <- X.createGC display window     
  X.setBackground display text_gc bg_color
  X.setForeground display text_gc fg_color

  pen_gc <- X.createGC display window     
  X.setBackground display pen_gc bg_color
  X.setForeground display pen_gc fg_color

  brush_gc <- X.createGC display window     
  X.setBackground display brush_gc bg_color
  X.setForeground display brush_gc fg_color

  redraw <- case redrawMode of
    Unbuffered -> return UnbufferedStuff
    DoubleBuffered -> do
      gc <- X.createGC display window
      X.setForeground display gc bg_color
      return (BufferedStuff gc depth)

  w <- newWindow display window text_gc pen_gc brush_gc ((x,y),(w,h)) redraw tickRate

  -- It might be some time till we get back to the event loop
  -- so we try to process as many events as possible now.
  -- This is a bit of a hack and partly aimed at avoiding the bug that
  -- directDraw might try to draw something before the first expose event
  -- is processed.
  -- To make the hack even more effective, we wait 10mS and synchronise
  -- before looking for the event.
  X.waitForEvent display 10
  X.sync display False
  minor_eloop display

  return w

closeWindow   = closeWindow' True
getWindowRect w = do
  mb_dc <- readIORef (ref_dc w)
  case mb_dc of 
    Just dc -> do 
       (pt,sz) <- readIORef (ref_rect dc)
       return (toPoint pt, toSize sz)
    Nothing -> 
       return ((0,0),(0,0)) -- ToDo?

-- main_eloop :: X.Display -> IO ()
-- main_eloop d = do
--   xevent <- X.allocXEvent
--   let loop = do
-- --    X.sync d False -- wild attempt to fix the broken X connection problem
--     count <- X.pending d
--     if (count > 0) then do
-- --      X.sync d False -- wild attempt to fix the broken X connection problem
-- 	 X.nextEvent d xevent
-- 	 window <- X.get_Window xevent
-- 	 wnd <- findWindow window
-- 	 etype <- X.get_EventType xevent
-- --      print (window,etype)
-- 	 dispatchEvent wnd etype xevent
-- 	 ws <- readIORef wnds
-- 	 unless (null ws) loop
-- 	else 
-- 	 loop
--   loop
--   X.free xevent               

-- This is the main event loop in the program
main_eloop :: X.Display -> IO ()
main_eloop d = do
  xevent <- X.allocXEvent
  let handleEvent = do
        count <- X.pending d
        next  <- T.nextTick timers
        if (count > 0 || not (isJust next)) 
         then do
           X.nextEvent d xevent
           window <- X.get_Window xevent
           etype  <- X.get_EventType xevent
--           print etype
           withWindow window $ \ wnd -> do
             dispatchEvent d wnd etype xevent
          else do
           timedOut <- X.waitForEvent d (fromJust next * 1000)
           T.fireTimers timers (fromJust next)
           return ()
  let loop = do
        ws <- readIORef wnds
        if (null ws) 
          then return ()
          else do
            handleEvent 
            loop
  loop
  X.free xevent               

-- This event loop is the same as above except that it is 
-- non-blocking: it only handles those events that have already arrived.
minor_eloop :: X.Display -> IO ()
minor_eloop d = do
  xevent <- X.allocXEvent
  let handleEvents = do
        count <- X.pending d
        if (count > 0) 
         then do
           X.nextEvent d xevent
           window <- X.get_Window xevent
           etype  <- X.get_EventType xevent
--           print etype
           withWindow window $ \ wnd -> do
             dispatchEvent d wnd etype xevent
           handleEvents                          -- loop!
          else do
           return ()
  handleEvents
  X.free xevent               

-- Time in milliseconds
getTime = X.gettimeofday_in_milliseconds

-- The DC is wrapped inside (IORef (Maybe ...)) so that we can mark
-- windows as being dead the moment they die and so that we don't
-- try to keep writing to them afterwards.
-- The events remain valid after the window dies.
-- It might be wiser to clear all events(???) and start returning
-- Closed whenever events are read - or (more GC friendly?), when
-- first read occurs but block thereafter?

data RedrawStuff 
  = UnbufferedStuff
  | BufferedStuff
      X.GC -- GC with foreground = background_color
      Int  -- depth

drawOnDC :: DC -> Draw () -> RedrawStuff -> IO ()
drawOnDC dc p redraw = 
  case redraw of
  UnbufferedStuff -> drawUnbuffered dc p
  BufferedStuff gc depth -> drawBuffered dc p gc depth

newWindow :: X.Display -> X.Window -> X.GC -> X.GC -> X.GC -> ((X.Position,X.Position),(X.Dimension,X.Dimension)) -> RedrawStuff -> Maybe T.Time -> IO Window
newWindow display window tgc pgc bgc rect redraw tickRate = do
  es  <- E.newEvents
  pic <- newIORef (return ())
-- failed attempts to find the default font
--  f'  <- X.fontFromGC display tgc
--  f   <- X.queryFont display f'
-- Since we can't ask the server what default font it chooses to bless
-- us with, we have to set an explicit font.  
  f   <- X.loadQueryFont display "9x15"  -- a random choice
  X.setFont display tgc (X.fontFromFontStruct f)
  bits <- newIORef DC_Bits
    { textColor = RGB 255 255 255
    , bkColor   = RGB 0   0   0
    , bkMode    = Transparent
    , textAlignment = (Left',Top)
    , brush     = RGB 255 255 255
    , pen       = defaultPen
    , font      = f
    }
  ref_rect <- newIORef rect
  dc     <- newIORef (Just MkDC{disp=display,drawable=window,textGC=tgc,paintGC=pgc,brushGC=bgc,ref_rect=ref_rect,ref_bits=bits})
  timer <- case tickRate of
   	   Just t  -> T.new timers t (E.sendTick es) >>= return.Just
   	   Nothing -> return Nothing
  let wnd = MkWindow{wnd=window,ref_dc=dc,events=es,graphic=pic,redraw=redraw,timer=timer}
  modIORef wnds (wnd:)
  return wnd

setGraphic w p = do 
  writeIORef (graphic w) p
  mb_dc <- readIORef (ref_dc w)
  case mb_dc of
   Just dc -> drawOnDC dc p (redraw w)
   Nothing -> return ()

getGraphic w = readIORef (graphic w)
modGraphic w = modIORef  (graphic w)

-- ToDo: according to X rules, we should not draw anything
-- until the first expose event arrives.
-- This bug is especially unfortunate when the drawing takes a long time
-- since we draw the picture twice in a row.
directDraw w p = do
  mb_dc <- readIORef (ref_dc w)
  case mb_dc of
   Just dc -> unDraw p dc
   Nothing -> return ()

findWindow :: X.Window -> IO Window
findWindow xw = do
  ws <- readIORef wnds
  return (head [ w | w <- ws, xw == wnd w ])  -- ToDo: don't use head

withWindow :: X.Window -> (Window -> IO ()) -> IO ()
withWindow xw k = do
  ws <- readIORef wnds
  case [ w | w <- ws, xw == wnd w ] of
   (w:_) -> k w
   _     -> return ()

send :: Window -> Event -> IO ()
send w e = E.sendEvent (events w) e

getWindowEvent w = E.getEvent (events w)

maybeGetWindowEvent w
  = do noEvent <- E.isNoEvent(events w)
       if noEvent 
          then return Nothing
          else do ev <- E.getEvent (events w)
                  return (Just ev)

dispatchEvent :: X.Display -> Window -> X.EventType -> X.XEventPtr -> IO ()
dispatchEvent display w etype xevent 
  | etype == X.graphicsExpose || etype == X.expose
  = paint
  | etype == X.motionNotify
  = mouseMove
  | etype == X.buttonPress
  = button True
  | etype == X.buttonRelease
  = button False
  | etype == X.keyPress
  = key True
  | etype == X.keyRelease
  = key False
  | etype == X.configureNotify
  = reconfig
  | etype == X.destroyNotify
  = destroy
  -- AC, 1/9/2000: treat a ClientMesage as a destroy event
  -- TODO: really need to examine the event in more detail,
  -- and ensure that xevent.xclient.message_type==ATOM_WM_PROTOCOLS &&
  --  xevent.xclient.data.l[0]==ATOM_WM_DELETE_WINDOW
  -- where ATOM_XXX is obtained from XInternAtom(dpy,"XXX",False)
  | etype == X.clientMessage
  = destroy

  -- ToDo: consider printing a warning message
  | otherwise
  = return ()

 where

  -- Redrawing is awkward because the request comes as a number of
  -- separate events.  We need to do one of the following (we currently
  -- do a combination of (1) and (3)):
  -- 1) Do a single redraw of the entire window but first delete all other
  --    expose events for this window from the queue.
  -- 2) Use all expose events for this window to build a Region object
  --    and use that to optimise redraws.
  -- 3) When double-buffering, use the buffer and information about
  --    whether it is up to date to serve redraws from the buffer.
  --    When single-buffering, use the server's backing store to reduce
  --    the number of expose events.  (Combine with bit-gravity info to
  --    handle resize requests.)
  paint :: IO ()
  paint = do
    let 
      stompOnExposeEvents = do
--        X.get_ExposeEvent xevent >>= print
        gotOne <- X.checkTypedWindowEvent display (wnd w) X.expose xevent
        if gotOne 
          then stompOnExposeEvents
          else return ()
    stompOnExposeEvents
    p <- readIORef (graphic w)
    mb_dc <- readIORef (ref_dc w)
    case mb_dc of 
     Just dc -> drawOnDC dc p (redraw w)
     Nothing -> return ()

  button :: Bool -> IO ()
  button isDown = do
    (_,_,_,x,y,_,_,_,b,_) <- X.get_ButtonEvent xevent
    let isLeft = b == 1 -- assume that button 1 = left button
    send w Button{pt = (x,y), isLeft=isLeft, isDown=isDown}

  key :: Bool -> IO ()
  key isDown = do
    sym <- X.lookupKeysym xevent 0
    send w Key{ char = toEnum (word32ToInt sym), isDown = isDown}

  mouseMove ::IO ()
  mouseMove = do
    (_,_,_,x,y,_,_,_,_,_) <- X.get_MotionEvent xevent
    send w MouseMove{ pt = (x,y) }

  reconfig :: IO ()
  reconfig = do
        (x,y,width,height) <- X.get_ConfigureEvent xevent
    	mb_dc <- readIORef (ref_dc w)
    	case mb_dc of 
    	  Just dc -> writeIORef (ref_rect dc) ((x,y),(width,height))
    	  Nothing -> return ()

	-- don't send new size, it may be out of date by the time we
	-- get round to reading the event
	send w Resize

  destroy :: IO ()
  destroy = do
	-- putStrLn "Window Destroyed" -- todo
        closeWindow' True w
        send w Closed

----------------------------------------------------------------
-- Utilities
----------------------------------------------------------------

modIORef :: IORef a -> (a -> a) -> IO ()
modIORef r f = do
  a <- readIORef r
  writeIORef r (f a)

----------------------------------------------------------------
-- End
----------------------------------------------------------------
