-----------------------------------------------------------------------------
--  $Id: ArrayBase.hs,v 1.13 2001/03/30 04:54:28 qrczak Exp $
--
-- (c) The GHC Team 2000
--
-- Basis for IArray and MArray.  Not intended for external consumption;
-- use IArray or MArray instead.

module ArrayBase where

import IOExts
import ST
import Addr
import Int
import Word

import List
import Ix
import qualified Array

import PrelST
import PrelWord
import PrelInt
import PrelStable
import PrelFloat
import PrelBase
import PrelPtr

-----------------------------------------------------------------------------
-- Class of immutable arrays:

class HasBounds a => IArray a e where
   array :: Ix ix => (ix,ix) -> [(ix,e)] -> a ix e
   (!)   :: Ix ix => a ix e -> ix -> e
   (//)  :: Ix ix => a ix e -> [(ix,e)] -> a ix e
   
   a // us = array (bounds a)
                   ([(i,a!i) | i <- indices a \\ [i | (i,_) <- us]] ++ us)

class HasBounds a where
   bounds :: Ix ix => a ix e -> (ix,ix)

-----------------------------------------------------------------------------
-- Normal polymorphic arrays

instance IArray Array.Array e where
   array = Array.array
   (!)   = (Array.!)
   (//)  = (Array.//)

instance HasBounds Array.Array where
   bounds = Array.bounds

-----------------------------------------------------------------------------
-- Flat unboxed arrays

data UArray ix a = UArray ix ix ByteArray#

instance HasBounds UArray where
   bounds (UArray l u _) = (l,u)

{-# INLINE newUArray #-}
newUArray :: (Ix ix, MArray (STUArray s) e (ST s))
          => (ix,ix) -> [(ix,e)] -> ST s (UArray ix e)
newUArray ixs ivs = do
    marr <- newArray_ ixs
    foldr (fill marr) (unsafeFreezeSTUArray marr) ivs

{-# INLINE fill #-}
fill :: (Ix ix, MArray (STUArray s) e (ST s))
     => STUArray s ix e -> (ix,e) -> ST s a -> ST s a
fill marr (i,v) next = writeArray marr i v >> next

unsafeFreezeSTUArray :: Ix ix => STUArray s ix e -> ST s (UArray ix e)
unsafeFreezeSTUArray (STUArray l u arr#) = ST $ \ s# ->
    case unsafeFreezeByteArray# arr# s# of { (# s2#, frozen #) ->
    (# s2#, UArray l u frozen #) }

{-# INLINE updateUArray #-}
updateUArray :: (MArray (STUArray s) e (ST s), Ix ix)
             => UArray ix e -> [(ix,e)] -> ST s (UArray ix e)
updateUArray a ivs = do
   marr <- thawSTUArray a
   foldr (fill marr) (unsafeFreezeSTUArray marr) ivs

showsUArray :: (IArray UArray e, Ix ix, Show ix, Show e)
            => Int -> UArray ix e -> ShowS
showsUArray p a = showParen (p > 9) $
                  showString "array " .
                  shows (bounds a) . showChar ' ' .
                  shows (assocs a)

-----------------------------------------------------------------------------
-- Flat unboxed arrays: instances

instance IArray UArray Bool where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexBoolArray
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Char where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexWideCharArray
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Int where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexIntArray
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Word where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexWordArray
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Addr where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexAddrArray
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray (Ptr a) where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexPtrArray
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Float where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexFloatArray
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Double where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexDoubleArray
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray (StablePtr a) where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexStablePtrArray
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Int8 where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexInt8Array
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Int16 where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexInt16Array
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Int32 where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexInt32Array
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Int64 where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexInt64Array
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Word8 where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexWord8Array
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Word16 where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexWord16Array
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Word32 where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexWord32Array
   a // ivs      = runST (updateUArray a ivs)

instance IArray UArray Word64 where
   array ixs ivs = runST (newUArray ixs ivs)
   (!)           = indexWord64Array
   a // ivs      = runST (updateUArray a ivs)

instance Ix ix => Eq (UArray ix Bool) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Char) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Int) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Word) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Addr) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix (Ptr a)) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Float) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Double) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix (StablePtr a)) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Int8) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Int16) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Int32) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Int64) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Word8) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Word16) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Word32) where
    a == a' = assocs a == assocs a'

instance Ix ix => Eq (UArray ix Word64) where
    a == a' = assocs a == assocs a'

instance Ix ix => Ord (UArray ix Bool) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Char) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Int) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Word) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Addr) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix (Ptr a)) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Float) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Double) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Int8) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Int16) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Int32) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Int64) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Word8) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Word16) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Word32) where
    compare a b = compare (assocs a) (assocs b)

instance Ix ix => Ord (UArray ix Word64) where
    compare a b = compare (assocs a) (assocs b)

instance (Ix ix, Show ix) => Show (UArray ix Bool) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Char) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Int) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Word) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Addr) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Float) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Double) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Int8) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Int16) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Int32) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Int64) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Word8) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Word16) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Word32) where
    showsPrec = showsUArray

instance (Ix ix, Show ix) => Show (UArray ix Word64) where
    showsPrec = showsUArray

-----------------------------------------------------------------------------
-- Mutable arrays

class (Monad m, HasBounds a) => MArray a e m where
   newArray   :: Ix ix => (ix,ix) -> e -> m (a ix e)
   newArray_  :: Ix ix => (ix,ix) -> m (a ix e)
   readArray  :: Ix ix => a ix e -> ix -> m e
   writeArray :: Ix ix => a ix e -> ix -> e -> m ()

   newArray ixs e
     = do a <- newArray_ ixs
          sequence_ [writeArray a i e | i <- range ixs]
          return a

   -- newArray takes an initialiser which all elements of
   -- the newly created array are initialised to.  newArray_ takes
   -- no initialiser, it is assumed that the array is initialised with
   -- "undefined" values.

   -- why not omit newArray_?  Because in the unboxed array case we would
   -- like to omit the initialisation altogether if possible.  We can't do
   -- this for boxed arrays, because the elements must all have valid values
   -- at all times in case of garbage collection.

   -- why not omit newArray?  Because in the boxed case, we can omit the
   -- default initialisation with undefined values if we *do* know the
   -- initial value and it is constant for all elements.

-----------------------------------------------------------------------------
-- Polymorphic non-strict mutable arrays (ST monad)

instance HasBounds (STArray s) where
   bounds = boundsSTArray

instance MArray (STArray s) e (ST s) where
   newArray      = newSTArray
   newArray_ ixs = newArray ixs st_bot
   readArray     = readSTArray
   writeArray    = writeSTArray

st_bot :: a
st_bot = error "STArray: undefined element"
        --ToDo: throw (ArrayException UndefinedElement)

-----------------------------------------------------------------------------
-- Flat unboxed mutable arrays (ST monad)

data STUArray s ix a = STUArray ix ix (MutableByteArray# s)

boundsSTUArray :: STUArray s ix a -> (ix,ix)
boundsSTUArray (STUArray l u _) = (l,u)

instance HasBounds (STUArray s) where
   bounds = boundsSTUArray

instance MArray (STUArray s) Bool (ST s) where
   newArray_  = newBoolArray
   readArray  = readBoolArray
   writeArray = writeBoolArray

instance MArray (STUArray s) Char (ST s) where
   newArray_  = newWideCharArray
   readArray  = readWideCharArray
   writeArray = writeWideCharArray

instance MArray (STUArray s) Int (ST s) where
   newArray_  = newIntArray
   readArray  = readIntArray
   writeArray = writeIntArray

instance MArray (STUArray s) Word (ST s) where
   newArray_  = newWordArray
   readArray  = readWordArray
   writeArray = writeWordArray

instance MArray (STUArray s) Addr (ST s) where
   newArray_  = newAddrArray
   readArray  = readAddrArray
   writeArray = writeAddrArray

instance MArray (STUArray s) (Ptr a) (ST s) where
   newArray_  = newPtrArray
   readArray  = readPtrArray
   writeArray = writePtrArray

instance MArray (STUArray s) Float (ST s) where
   newArray_  = newFloatArray
   readArray  = readFloatArray
   writeArray = writeFloatArray

instance MArray (STUArray s) Double (ST s) where
   newArray_  = newDoubleArray
   readArray  = readDoubleArray
   writeArray = writeDoubleArray

instance MArray (STUArray s) (StablePtr a) (ST s) where
   newArray_  = newStablePtrArray
   readArray  = readStablePtrArray
   writeArray = writeStablePtrArray

instance MArray (STUArray s) Int8 (ST s) where
   newArray_  = newInt8Array
   readArray  = readInt8Array
   writeArray = writeInt8Array

instance MArray (STUArray s) Int16 (ST s) where
   newArray_  = newInt16Array
   readArray  = readInt16Array
   writeArray = writeInt16Array

instance MArray (STUArray s) Int32 (ST s) where
   newArray_  = newInt32Array
   readArray  = readInt32Array
   writeArray = writeInt32Array

instance MArray (STUArray s) Int64 (ST s) where
   newArray_  = newInt64Array
   readArray  = readInt64Array
   writeArray = writeInt64Array

instance MArray (STUArray s) Word8 (ST s) where
   newArray_  = newWord8Array
   readArray  = readWord8Array
   writeArray = writeWord8Array

instance MArray (STUArray s) Word16 (ST s) where
   newArray_  = newWord16Array
   readArray  = readWord16Array
   writeArray = writeWord16Array

instance MArray (STUArray s) Word32 (ST s) where
   newArray_  = newWord32Array
   readArray  = readWord32Array
   writeArray = writeWord32Array

instance MArray (STUArray s) Word64 (ST s) where
   newArray_  = newWord64Array
   readArray  = readWord64Array
   writeArray = writeWord64Array

-----------------------------------------------------------------------------
-- Low-level operations: firstly, creating new mutable arrays

newBoolArray      :: Ix ix => (ix,ix) -> ST s (STUArray s ix Bool)
newCharArray      :: Ix ix => (ix,ix) -> ST s (STUArray s ix Char)
newWideCharArray  :: Ix ix => (ix,ix) -> ST s (STUArray s ix Char)
newIntArray       :: Ix ix => (ix,ix) -> ST s (STUArray s ix Int)
newWordArray      :: Ix ix => (ix,ix) -> ST s (STUArray s ix Word)
newAddrArray      :: Ix ix => (ix,ix) -> ST s (STUArray s ix Addr)
newPtrArray       :: Ix ix => (ix,ix) -> ST s (STUArray s ix (Ptr a))
newFloatArray     :: Ix ix => (ix,ix) -> ST s (STUArray s ix Float)
newDoubleArray    :: Ix ix => (ix,ix) -> ST s (STUArray s ix Double)
newStablePtrArray :: Ix ix => (ix,ix) -> ST s (STUArray s ix (StablePtr a))
newInt8Array      :: Ix ix => (ix,ix) -> ST s (STUArray s ix Int8)
newInt16Array     :: Ix ix => (ix,ix) -> ST s (STUArray s ix Int16)
newInt32Array     :: Ix ix => (ix,ix) -> ST s (STUArray s ix Int32)
newInt64Array     :: Ix ix => (ix,ix) -> ST s (STUArray s ix Int64)
newWord8Array     :: Ix ix => (ix,ix) -> ST s (STUArray s ix Word8)
newWord16Array    :: Ix ix => (ix,ix) -> ST s (STUArray s ix Word16)
newWord32Array    :: Ix ix => (ix,ix) -> ST s (STUArray s ix Word32)
newWord64Array    :: Ix ix => (ix,ix) -> ST s (STUArray s ix Word64)

type IPr = (Int, Int)
{-# SPECIALIZE newBoolArray      :: IPr -> ST s (STUArray s Int Bool)          #-}
{-# SPECIALIZE newCharArray      :: IPr -> ST s (STUArray s Int Char)          #-}
{-# SPECIALIZE newWideCharArray  :: IPr -> ST s (STUArray s Int Char)          #-}
{-# SPECIALIZE newIntArray       :: IPr -> ST s (STUArray s Int Int)           #-}
{-# SPECIALIZE newWordArray      :: IPr -> ST s (STUArray s Int Word)          #-}
{-# SPECIALIZE newAddrArray      :: IPr -> ST s (STUArray s Int Addr)          #-}
{-# SPECIALIZE newPtrArray       :: IPr -> ST s (STUArray s Int (Ptr a))       #-}
{-# SPECIALIZE newFloatArray     :: IPr -> ST s (STUArray s Int Float)         #-}
{-# SPECIALIZE newDoubleArray    :: IPr -> ST s (STUArray s Int Double)        #-}
{-# SPECIALIZE newStablePtrArray :: IPr -> ST s (STUArray s Int (StablePtr a)) #-}
{-# SPECIALIZE newInt8Array      :: IPr -> ST s (STUArray s Int Int8)          #-}
{-# SPECIALIZE newInt16Array     :: IPr -> ST s (STUArray s Int Int16)         #-}
{-# SPECIALIZE newInt32Array     :: IPr -> ST s (STUArray s Int Int32)         #-}
{-# SPECIALIZE newInt64Array     :: IPr -> ST s (STUArray s Int Int64)         #-}
{-# SPECIALIZE newWord8Array     :: IPr -> ST s (STUArray s Int Word8)         #-}
{-# SPECIALIZE newWord16Array    :: IPr -> ST s (STUArray s Int Word16)        #-}
{-# SPECIALIZE newWord32Array    :: IPr -> ST s (STUArray s Int Word32)        #-}
{-# SPECIALIZE newWord64Array    :: IPr -> ST s (STUArray s Int Word64)        #-}

newBoolArray (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (bOOL_SCALE n#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newCharArray (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# n# s#         of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newWideCharArray (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (n# *# 4#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newIntArray (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (wORD_SCALE n#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newWordArray (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (wORD_SCALE n#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newAddrArray (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (wORD_SCALE n#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newPtrArray (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (wORD_SCALE n#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newFloatArray (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (fLOAT_SCALE n#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newDoubleArray (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (dOUBLE_SCALE n#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newStablePtrArray ixs@(l,u) = ST $ \ s# ->
    case rangeSize ixs               of { I# n# ->
    case newByteArray# (wORD_SCALE n#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newInt8Array (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# n# s#         of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newInt16Array (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (n# *# 2#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newInt32Array (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (n# *# 4#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newInt64Array (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (n# *# 8#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newWord8Array (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# n# s#         of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newWord16Array (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (n# *# 2#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newWord32Array (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (n# *# 4#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

newWord64Array (l,u) = ST $ \ s# ->
    case rangeSize (l,u)             of { I# n# ->
    case newByteArray# (n# *# 8#) s# of { (# s2#, barr# #) ->
    (# s2#, STUArray l u barr# #) }}

#include "config.h"

bOOL_INDEX, bOOL_SCALE, wORD_SCALE, dOUBLE_SCALE, fLOAT_SCALE :: Int# -> Int#
bOOL_INDEX   n# = n# `iShiftRA#` scaleBits#
#if SIZEOF_VOID_P == 4
    where scaleBits# = 5#
#else
    where scaleBits# = 6#
#endif
bOOL_SCALE   n# = bOOL_INDEX (n# +# last#) where I# last# = SIZEOF_VOID_P - 1
wORD_SCALE   n# = scale# *# n# where I# scale# = SIZEOF_VOID_P
dOUBLE_SCALE n# = scale# *# n# where I# scale# = SIZEOF_DOUBLE
fLOAT_SCALE  n# = scale# *# n# where I# scale# = SIZEOF_FLOAT

-----------------------------------------------------------------------------
-- reading from mutable unboxed arrays

readBoolArray      :: Ix ix => STUArray s ix Bool          -> ix -> ST s Bool
readCharArray      :: Ix ix => STUArray s ix Char          -> ix -> ST s Char
readWideCharArray  :: Ix ix => STUArray s ix Char          -> ix -> ST s Char
readIntArray       :: Ix ix => STUArray s ix Int           -> ix -> ST s Int
readWordArray      :: Ix ix => STUArray s ix Word          -> ix -> ST s Word
readAddrArray      :: Ix ix => STUArray s ix Addr          -> ix -> ST s Addr
readPtrArray       :: Ix ix => STUArray s ix (Ptr a)       -> ix -> ST s (Ptr a)
readFloatArray     :: Ix ix => STUArray s ix Float         -> ix -> ST s Float
readDoubleArray    :: Ix ix => STUArray s ix Double        -> ix -> ST s Double
readStablePtrArray :: Ix ix => STUArray s ix (StablePtr a) -> ix -> ST s (StablePtr a)
readInt8Array      :: Ix ix => STUArray s ix Int8          -> ix -> ST s Int8
readInt16Array     :: Ix ix => STUArray s ix Int16         -> ix -> ST s Int16
readInt32Array     :: Ix ix => STUArray s ix Int32         -> ix -> ST s Int32
readInt64Array     :: Ix ix => STUArray s ix Int64         -> ix -> ST s Int64
readWord8Array     :: Ix ix => STUArray s ix Word8         -> ix -> ST s Word8
readWord16Array    :: Ix ix => STUArray s ix Word16        -> ix -> ST s Word16
readWord32Array    :: Ix ix => STUArray s ix Word32        -> ix -> ST s Word32
readWord64Array    :: Ix ix => STUArray s ix Word64        -> ix -> ST s Word64

{-# SPECIALIZE readBoolArray      :: STUArray s Int Bool          -> Int -> ST s Bool          #-}
{-# SPECIALIZE readCharArray      :: STUArray s Int Char          -> Int -> ST s Char          #-}
{-# SPECIALIZE readWideCharArray  :: STUArray s Int Char          -> Int -> ST s Char          #-}
{-# SPECIALIZE readIntArray       :: STUArray s Int Int           -> Int -> ST s Int           #-}
{-# SPECIALIZE readWordArray      :: STUArray s Int Word          -> Int -> ST s Word          #-}
{-# SPECIALIZE readAddrArray      :: STUArray s Int Addr          -> Int -> ST s Addr          #-}
{-# SPECIALIZE readPtrArray       :: STUArray s Int (Ptr a)       -> Int -> ST s (Ptr a)       #-}
{-# SPECIALIZE readFloatArray     :: STUArray s Int Float         -> Int -> ST s Float         #-}
{-# SPECIALIZE readDoubleArray    :: STUArray s Int Double        -> Int -> ST s Double        #-}
{-# SPECIALIZE readStablePtrArray :: STUArray s Int (StablePtr a) -> Int -> ST s (StablePtr a) #-}
{-# SPECIALIZE readInt8Array      :: STUArray s Int Int8          -> Int -> ST s Int8          #-}
{-# SPECIALIZE readInt16Array     :: STUArray s Int Int16         -> Int -> ST s Int16         #-}
{-# SPECIALIZE readInt32Array     :: STUArray s Int Int32         -> Int -> ST s Int32         #-}
{-# SPECIALIZE readInt64Array     :: STUArray s Int Int64         -> Int -> ST s Int64         #-}
{-# SPECIALIZE readWord8Array     :: STUArray s Int Word8         -> Int -> ST s Word8         #-}
{-# SPECIALIZE readWord16Array    :: STUArray s Int Word16        -> Int -> ST s Word16        #-}
{-# SPECIALIZE readWord32Array    :: STUArray s Int Word32        -> Int -> ST s Word32        #-}
{-# SPECIALIZE readWord64Array    :: STUArray s Int Word64        -> Int -> ST s Word64        #-}

readBoolArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readWordArray# barr# (bOOL_INDEX n#) s# of { (# s2#, r# #) ->
    (# s2#, (r# `and#` bOOL_BIT n#) `neWord#` int2Word# 0# #) }}

readCharArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readCharArray# barr# n# s#      of { (# s2#, r# #) ->
    (# s2#, C# r# #) }}

readWideCharArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readWideCharArray# barr# n# s#  of { (# s2#, r# #) ->
    (# s2#, C# r# #) }}

readIntArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readIntArray# barr# n# s#       of { (# s2#, r# #) ->
    (# s2#, I# r# #) }}

readWordArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readWordArray# barr# n# s#      of { (# s2#, r# #) ->
    (# s2#, W# r# #) }}

readAddrArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readAddrArray# barr# n# s#      of { (# s2#, r# #) ->
    (# s2#, A# r# #) }}

readPtrArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readAddrArray# barr# n# s#      of { (# s2#, r# #) ->
    (# s2#, Ptr r# #) }}

readFloatArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readFloatArray# barr# n# s#     of { (# s2#, r# #) ->
    (# s2#, F# r# #) }}

readDoubleArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readDoubleArray# barr# n# s#    of { (# s2#, r# #) ->
    (# s2#, D# r# #) }}

readStablePtrArray (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readStablePtrArray# barr# n# s# of { (# s2#, r# #) ->
    (# s2# , StablePtr r# #) }}

readInt8Array (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readInt8Array# barr# n# s#      of { (# s2#, r# #) ->
    (# s2#, I8# r# #) }}

readInt16Array (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readInt16Array# barr# n# s#     of { (# s2#, r# #) ->
    (# s2#, I16# r# #) }}

readInt32Array (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readInt32Array# barr# n# s#     of { (# s2#, r# #) ->
    (# s2#, I32# r# #) }}

readInt64Array (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readInt64Array# barr# n# s#     of { (# s2#, r# #) ->
    (# s2#, I64# r# #) }}

readWord8Array (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readWord8Array# barr# n# s#     of { (# s2#, r# #) ->
    (# s2#, W8# r# #) }}

readWord16Array (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readWord16Array# barr# n# s#    of { (# s2#, r# #) ->
    (# s2#, W16# r# #) }}

readWord32Array (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readWord32Array# barr# n# s#    of { (# s2#, r# #) ->
    (# s2#, W32# r# #) }}

readWord64Array (STUArray l u barr#) n = ST $ \ s# ->
    case index (l,u) n                   of { I# n# ->
    case readWord64Array# barr# n# s#    of { (# s2#, r# #) ->
    (# s2#, W64# r# #) }}

bOOL_BIT, bOOL_NOT_BIT :: Int# -> Word#
bOOL_BIT     n# = int2Word# 1# `shiftL#` (word2Int# (int2Word# n# `and#` mask#))
    where W# mask# = SIZEOF_VOID_P * 8 - 1
bOOL_NOT_BIT n# = bOOL_BIT n# `xor#` mb# where W# mb# = maxBound

-----------------------------------------------------------------------------
-- writing into mutable unboxed arrays

writeBoolArray      :: Ix ix => STUArray s ix Bool          -> ix -> Bool        -> ST s ()
writeCharArray      :: Ix ix => STUArray s ix Char          -> ix -> Char        -> ST s ()
writeWideCharArray  :: Ix ix => STUArray s ix Char          -> ix -> Char        -> ST s ()
writeIntArray       :: Ix ix => STUArray s ix Int           -> ix -> Int         -> ST s ()
writeWordArray      :: Ix ix => STUArray s ix Word          -> ix -> Word        -> ST s ()
writeAddrArray      :: Ix ix => STUArray s ix Addr          -> ix -> Addr        -> ST s ()
writePtrArray       :: Ix ix => STUArray s ix (Ptr a)       -> ix -> Ptr a       -> ST s ()
writeFloatArray     :: Ix ix => STUArray s ix Float         -> ix -> Float       -> ST s ()
writeDoubleArray    :: Ix ix => STUArray s ix Double        -> ix -> Double      -> ST s ()
writeStablePtrArray :: Ix ix => STUArray s ix (StablePtr a) -> ix -> StablePtr a -> ST s ()
writeInt8Array      :: Ix ix => STUArray s ix Int8          -> ix -> Int8        -> ST s ()
writeInt16Array     :: Ix ix => STUArray s ix Int16         -> ix -> Int16       -> ST s ()
writeInt32Array     :: Ix ix => STUArray s ix Int32         -> ix -> Int32       -> ST s ()
writeInt64Array     :: Ix ix => STUArray s ix Int64         -> ix -> Int64       -> ST s ()
writeWord8Array     :: Ix ix => STUArray s ix Word8         -> ix -> Word8       -> ST s ()
writeWord16Array    :: Ix ix => STUArray s ix Word16        -> ix -> Word16      -> ST s ()
writeWord32Array    :: Ix ix => STUArray s ix Word32        -> ix -> Word32      -> ST s ()
writeWord64Array    :: Ix ix => STUArray s ix Word64        -> ix -> Word64      -> ST s ()

{-# SPECIALIZE writeBoolArray      :: STUArray s Int Bool          -> Int -> Bool        -> ST s () #-}
{-# SPECIALIZE writeCharArray      :: STUArray s Int Char          -> Int -> Char        -> ST s () #-}
{-# SPECIALIZE writeWideCharArray  :: STUArray s Int Char          -> Int -> Char        -> ST s () #-}
{-# SPECIALIZE writeIntArray       :: STUArray s Int Int           -> Int -> Int         -> ST s () #-}
{-# SPECIALIZE writeWordArray      :: STUArray s Int Word          -> Int -> Word        -> ST s () #-}
{-# SPECIALIZE writeAddrArray      :: STUArray s Int Addr          -> Int -> Addr        -> ST s () #-}
{-# SPECIALIZE writePtrArray       :: STUArray s Int (Ptr a)       -> Int -> (Ptr a)     -> ST s () #-}
{-# SPECIALIZE writeFloatArray     :: STUArray s Int Float         -> Int -> Float       -> ST s () #-}
{-# SPECIALIZE writeDoubleArray    :: STUArray s Int Double        -> Int -> Double      -> ST s () #-}
{-# SPECIALIZE writeStablePtrArray :: STUArray s Int (StablePtr a) -> Int -> StablePtr a -> ST s () #-}
{-# SPECIALIZE writeInt8Array      :: STUArray s Int Int8          -> Int -> Int8        -> ST s () #-}
{-# SPECIALIZE writeInt16Array     :: STUArray s Int Int16         -> Int -> Int16       -> ST s () #-}
{-# SPECIALIZE writeInt32Array     :: STUArray s Int Int32         -> Int -> Int32       -> ST s () #-}
{-# SPECIALIZE writeInt64Array     :: STUArray s Int Int64         -> Int -> Int64       -> ST s () #-}
{-# SPECIALIZE writeWord8Array     :: STUArray s Int Word8         -> Int -> Word8       -> ST s () #-}
{-# SPECIALIZE writeWord16Array    :: STUArray s Int Word16        -> Int -> Word16      -> ST s () #-}
{-# SPECIALIZE writeWord32Array    :: STUArray s Int Word32        -> Int -> Word32      -> ST s () #-}
{-# SPECIALIZE writeWord64Array    :: STUArray s Int Word64        -> Int -> Word64      -> ST s () #-}

writeBoolArray (STUArray l u barr#) n ele = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case bOOL_INDEX n#                        of { m# ->
    case readWordArray# barr# m# s#           of { (# s2#, old# #) ->
    case if ele then old# `or#` bOOL_BIT n#
         else old# `and#` bOOL_NOT_BIT n#     of { r# ->
    case writeWordArray# barr# m# r# s#       of { s2# ->
    (# s2#, () #) }}}}}

writeCharArray (STUArray l u barr#) n (C# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeCharArray# barr# n# ele s#      of { s2#   ->
    (# s2#, () #) }}

writeWideCharArray (STUArray l u barr#) n (C# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeWideCharArray# barr# n# ele s#  of { s2#   ->
    (# s2#, () #) }}

writeIntArray (STUArray l u barr#) n (I# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeIntArray# barr# n# ele s#       of { s2#   ->
    (# s2#, () #) }}

writeWordArray (STUArray l u barr#) n (W# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeWordArray# barr# n# ele s#      of { s2#   ->
    (# s2#, () #) }}

writeAddrArray (STUArray l u barr#) n (A# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeAddrArray# barr# n# ele s#      of { s2#   ->
    (# s2#, () #) }}

writePtrArray (STUArray l u barr#) n (Ptr ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeAddrArray# barr# n# ele s#      of { s2#   ->
    (# s2#, () #) }}

writeFloatArray (STUArray l u barr#) n (F# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeFloatArray# barr# n# ele s#     of { s2#   ->
    (# s2#, () #) }}

writeDoubleArray (STUArray l u barr#) n (D# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeDoubleArray# barr# n# ele s#    of { s2#   ->
    (# s2#, () #) }}

writeStablePtrArray (STUArray l u barr#) n (StablePtr sp#) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeStablePtrArray# barr# n# sp# s# of { s2#   ->
    (# s2# , () #) }}

writeInt8Array (STUArray l u barr#) n (I8# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeInt8Array# barr# n# ele s#      of { s2#   ->
    (# s2#, () #) }}

writeInt16Array (STUArray l u barr#) n (I16# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeInt16Array# barr# n# ele s#     of { s2#   ->
    (# s2#, () #) }}

writeInt32Array (STUArray l u barr#) n (I32# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeInt32Array# barr# n# ele s#     of { s2#   ->
    (# s2#, () #) }}

writeInt64Array (STUArray l u barr#) n (I64# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeInt64Array# barr# n# ele s#     of { s2#   ->
    (# s2#, () #) }}

writeWord8Array (STUArray l u barr#) n (W8# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeWord8Array# barr# n# ele s#     of { s2#   ->
    (# s2#, () #) }}

writeWord16Array (STUArray l u barr#) n (W16# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeWord16Array# barr# n# ele s#    of { s2#   ->
    (# s2#, () #) }}

writeWord32Array (STUArray l u barr#) n (W32# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeWord32Array# barr# n# ele s#    of { s2#   ->
    (# s2#, () #) }}

writeWord64Array (STUArray l u barr#) n (W64# ele) = ST $ \ s# ->
    case index (l,u) n                        of { I# n# ->
    case writeWord64Array# barr# n# ele s#    of { s2#   ->
    (# s2#, () #) }}

-----------------------------------------------------------------------------
-- indexing into immutable unboxed arrays

indexBoolArray      :: Ix ix => UArray ix Bool          -> ix -> Bool
indexCharArray      :: Ix ix => UArray ix Char          -> ix -> Char
indexWideCharArray  :: Ix ix => UArray ix Char          -> ix -> Char
indexIntArray       :: Ix ix => UArray ix Int           -> ix -> Int
indexWordArray      :: Ix ix => UArray ix Word          -> ix -> Word
indexAddrArray      :: Ix ix => UArray ix Addr          -> ix -> Addr
indexPtrArray       :: Ix ix => UArray ix (Ptr a)       -> ix -> Ptr a
indexFloatArray     :: Ix ix => UArray ix Float         -> ix -> Float
indexDoubleArray    :: Ix ix => UArray ix Double        -> ix -> Double
indexStablePtrArray :: Ix ix => UArray ix (StablePtr a) -> ix -> StablePtr a
indexInt8Array      :: Ix ix => UArray ix Int8          -> ix -> Int8
indexInt16Array     :: Ix ix => UArray ix Int16         -> ix -> Int16
indexInt32Array     :: Ix ix => UArray ix Int32         -> ix -> Int32
indexInt64Array     :: Ix ix => UArray ix Int64         -> ix -> Int64
indexWord8Array     :: Ix ix => UArray ix Word8         -> ix -> Word8
indexWord16Array    :: Ix ix => UArray ix Word16        -> ix -> Word16
indexWord32Array    :: Ix ix => UArray ix Word32        -> ix -> Word32
indexWord64Array    :: Ix ix => UArray ix Word64        -> ix -> Word64

{-# SPECIALIZE indexBoolArray      :: UArray Int Bool          -> Int -> Bool          #-}
{-# SPECIALIZE indexCharArray      :: UArray Int Char          -> Int -> Char          #-}
{-# SPECIALIZE indexWideCharArray  :: UArray Int Char          -> Int -> Char          #-}
{-# SPECIALIZE indexIntArray       :: UArray Int Int           -> Int -> Int           #-}
{-# SPECIALIZE indexWordArray      :: UArray Int Word          -> Int -> Word          #-}
{-# SPECIALIZE indexAddrArray      :: UArray Int Addr          -> Int -> Addr          #-}
{-# SPECIALIZE indexPtrArray       :: UArray Int (Ptr a)       -> Int -> (Ptr a)       #-}
{-# SPECIALIZE indexFloatArray     :: UArray Int Float         -> Int -> Float         #-}
{-# SPECIALIZE indexDoubleArray    :: UArray Int Double        -> Int -> Double        #-}
{-# SPECIALIZE indexStablePtrArray :: UArray Int (StablePtr a) -> Int -> (StablePtr a) #-}
{-# SPECIALIZE indexInt8Array      :: UArray Int Int8          -> Int -> Int8          #-}
{-# SPECIALIZE indexInt16Array     :: UArray Int Int16         -> Int -> Int16         #-}
{-# SPECIALIZE indexInt32Array     :: UArray Int Int32         -> Int -> Int32         #-}
{-# SPECIALIZE indexInt64Array     :: UArray Int Int64         -> Int -> Int64         #-}
{-# SPECIALIZE indexWord8Array     :: UArray Int Word8         -> Int -> Word8         #-}
{-# SPECIALIZE indexWord16Array    :: UArray Int Word16        -> Int -> Word16        #-}
{-# SPECIALIZE indexWord32Array    :: UArray Int Word32        -> Int -> Word32        #-}
{-# SPECIALIZE indexWord64Array    :: UArray Int Word64        -> Int -> Word64        #-}

indexBoolArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexWordArray# barr# (bOOL_INDEX n#) of { r# ->
    (r# `and#` bOOL_BIT n#) `neWord#` int2Word# 0# }}

indexCharArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexCharArray# barr# n#      of { r# ->
    (C# r#) }}

indexWideCharArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexWideCharArray# barr# n#  of { r# ->
    (C# r#) }}

indexIntArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexIntArray# barr# n#       of { r# ->
    (I# r#) }}

indexWordArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexWordArray# barr# n#      of { r# ->
    (W# r#) }}

indexAddrArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexAddrArray# barr# n#      of { r# ->
    (A# r#) }}

indexPtrArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexAddrArray# barr# n#      of { r# ->
    (Ptr r#) }}

indexFloatArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexFloatArray# barr# n#     of { r# ->
    (F# r#) }}

indexDoubleArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexDoubleArray# barr# n#    of { r# ->
    (D# r#) }}

indexStablePtrArray (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexStablePtrArray# barr# n# of { r# ->
    (StablePtr r#) }}

indexInt8Array (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexInt8Array# barr# n#      of { r# ->
    (I8# r#) }}

indexInt16Array (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexInt16Array# barr# n#     of { r# ->
    (I16# r#) }}

indexInt32Array (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexInt32Array# barr# n#     of { r# ->
    (I32# r#) }}

indexInt64Array (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexInt64Array# barr# n#     of { r# ->
    (I64# r#) }}

indexWord8Array (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexWord8Array# barr# n#     of { r# ->
    (W8# r#) }}

indexWord16Array (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexWord16Array# barr# n#    of { r# ->
    (W16# r#) }}

indexWord32Array (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexWord32Array# barr# n#    of { r# ->
    (W32# r#) }}

indexWord64Array (UArray l u barr#) n
  = case index (l,u) n                 of { I# n# ->
    case indexWord64Array# barr# n#    of { r# ->
    (W64# r#) }}

-----------------------------------------------------------------------------
-- Freezing

freeze  :: (Ix ix, MArray a e m, IArray b e) => a ix e -> m (b ix e)
freeze a = do { elts <- getAssocs a; return (array (bounds a) elts) }

freezeSTUArray   :: Ix ix => STUArray s ix e -> ST s (UArray ix e)
-- This coercion of memcpy to the ST monad is safe, because memcpy
-- only modifies its destination operand, which is already MutableByteArray.
freezeSTUArray (STUArray l u arr) = ST $ \ s ->
        let n = sizeofMutableByteArray# arr in
        case (newByteArray# n s)                   of { (# s, newarr #) ->
        case ((unsafeCoerce# memcpy) newarr arr n s) of { (# s, () #) ->
        case unsafeFreezeByteArray# newarr s       of { (# s, frozen #) ->
        (# s, UArray l u frozen #) }}}

foreign import "memcpy" unsafe
  memcpy :: MutableByteArray# RealWorld -> ByteArray# -> Int# -> IO ()

{-# RULES
"freeze/STArray"    freeze = freezeSTArray
"freeze/IOArray"    freeze = freezeIOArray
"freeze/STUArray"   freeze = freezeSTUArray
 #-}

unsafeFreeze :: (Ix ix, MArray a e m, IArray b e) => a ix e -> m (b ix e)
unsafeFreeze = freeze

{-# RULES
"unsafeFreeze/STArray"    unsafeFreeze = unsafeFreezeSTArray
"unsafeFreeze/IOArray"    unsafeFreeze = unsafeFreezeIOArray
"unsafeFreeze/STUArray"   unsafeFreeze = unsafeFreezeSTUArray
 #-}

-----------------------------------------------------------------------------
-- Thawing

thaw :: (Ix ix, IArray a e, MArray b e m) => a ix e -> m (b ix e)
thaw a = do
  new_a <- newArray_ (bounds a)
  sequence_ [ writeArray new_a i e | (i,e) <- assocs a ]
  return new_a

thawSTUArray :: Ix ix => UArray ix e -> ST s (STUArray s ix e)
thawSTUArray (UArray l u arr) = ST $ \ s ->
        let n = sizeofByteArray# arr in
        case (newByteArray# n s)                   of { (# s, newarr #) ->
        case ((unsafeCoerce# memcpy) newarr arr n s) of { (# s, () #) ->
        (# s, STUArray l u newarr #) }}

{-# RULES
"thaw/STArray"    thaw = thawSTArray
"thaw/IOArray"    thaw = thawIOArray
"thaw/STUArray"   thaw = thawSTUArray
 #-}

{-
  in-place conversion of immutable arrays to mutable ones places
  a proof obligation on the user: no other parts of your code can
  have a reference to the array at the point where you unsafely
  thaw it (and, subsequently mutate it, I suspect.)
-}
unsafeThaw :: (Ix ix, IArray a e, MArray b e m) => a ix e -> m (b ix e)
unsafeThaw = thaw

unsafeThawSTUArray :: Ix ix => UArray ix e -> ST s (STUArray s ix e)
unsafeThawSTUArray (UArray l u barr)
  = return (STUArray l u (unsafeCoerce# barr))

{-# RULES
"unsafeThaw/STArray"    unsafeThaw = unsafeThawSTArray
"unsafeThaw/IOArray"    unsafeThaw = unsafeThawIOArray
"unsafeThaw/STUArray"   unsafeThaw = unsafeThawSTUArray
 #-}

-----------------------------------------------------------------------------
-- utils

{-# INLINE assocs #-}
assocs   :: (Ix ix, IArray a e) => a ix e -> [(ix,e)]
assocs a =  [(i, a!i) | i <- indices a]

{-# INLINE getAssocs #-}
getAssocs   :: (Ix ix, MArray a e m) => a ix e -> m [(ix,e)]
getAssocs a =  sequence [ do { e <- readArray a i; return (i,e) }
                        | i <- indices a ]

{-# INLINE indices #-}
indices :: (Ix ix, HasBounds a) => a ix e -> [ix]
indices =  range . bounds
