------------------------------------------------------------------------
-- $Id: DiffArray.hs,v 1.2 2001/03/31 10:04:13 qrczak Exp $
--
-- (c) The GHC Team 2001
--

module DiffArray (

    -- Diff arrays have immutable interface, but rely on internal
    -- updates in place to provide fast functional update operator
    -- '//'.
    --
    -- When the '//' operator is applied to a diff array, its contents
    -- are physically updated in place. The old array silently changes
    -- its representation without changing the visible behavior:
    -- it stores a link to the new current array along with the
    -- difference to be applied to get the old contents.
    --
    -- So if a diff array is used in a single-threaded style,
    -- i.e. after '//' application the old version is no longer used,
    -- 'a!i' takes O(1) time and 'a // d' takes O(length d). Accessing
    -- elements of older versions gradually becomes slower.
    --
    -- Updating an array which is not current makes a physical copy.
    -- The resulting array is unlinked from the old family. So you
    -- can obtain a version which is guaranteed to be current and
    -- thus have fast element access by 'a // []'.

    -- Possible improvement for the future (not implemented now):
    -- make it possible to say "I will make an update now, but when
    -- I later return to the old version, I want it to mutate back
    -- instead of being copied".

    -- An arbitrary MArray type living in the IO monad can be converted
    -- to a diff array.
    IOToDiffArray, -- data IOToDiffArray
                   --     (a :: * -> * -> *) -- internal mutable array
                   --     (i :: *)           -- indices
                   --     (e :: *)           -- elements

    -- Two most important diff array types are fully polymorphic
    -- lazy boxed DiffArray:
    DiffArray,     -- = IOToDiffArray IOArray
    -- ...and strict unboxed DiffUArray, working only for elements
    -- of primitive types but more compact and usually faster:
    DiffUArray,    -- = IOToDiffArray IOUArray
    
    -- Module IArray provides the interface of diff arrays. They are
    -- instances of class IArray.
    module IArray,
    
    -- These are really internal functions, but you will need them
    -- to make further IArray instances of various DiffArrays (for
    -- either more MArray types or more unboxed element types).
    newDiffArray, indexDiffArray, updateDiffArray
    )
    where

------------------------------------------------------------------------
-- Imports.

import IArray
import MArray

import Word          (Word)
import Addr          (Addr)
import Ptr           (Ptr)
import StablePtr     (StablePtr)
import Int           (Int8,  Int16,  Int32,  Int64)
import Word          (Word8, Word16, Word32, Word64)

import IOExts        (unsafePerformIO)
import PrelConc      (MVar, newMVar, takeMVar, putMVar)
import PrelException (block)

------------------------------------------------------------------------
-- Diff array types.

-- Convert an IO array type to a diff array.
newtype IOToDiffArray a i e =
    DiffArray {varDiffArray :: MVar (DiffArrayData a i e)}

-- Internal representation: either a mutable array, or a link to
-- another diff array patched with a list of index+element pairs.
data DiffArrayData a i e = Current (a i e)
                         | Diff (IOToDiffArray a i e) [(i, e)]

-- Type synonyms for two most important IO array types.
type DiffArray  = IOToDiffArray IOArray
type DiffUArray = IOToDiffArray IOUArray

-- Having 'MArray a e IO' in instance context would require
-- -fallow-undecidable-instances, so each instance is separate here.

------------------------------------------------------------------------
-- Boring instances.

instance IArray (IOToDiffArray IOArray) e where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Char where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Int where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Word where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Addr where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) (Ptr a) where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Float where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Double where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) (StablePtr a) where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Int8 where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Int16 where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Int32 where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Int64 where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Word8 where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Word16 where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Word32 where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance IArray (IOToDiffArray IOUArray) Word64 where
    array is ies = unsafePerformIO $ newDiffArray is ies
    a!i          = unsafePerformIO $ a `indexDiffArray` i
    a // ies     = unsafePerformIO $ a `updateDiffArray` ies

instance HasBounds a => HasBounds (IOToDiffArray a) where
    bounds a = unsafePerformIO $ boundsDiffArray a

------------------------------------------------------------------------
-- The important stuff.

newDiffArray :: (MArray a e IO, Ix ix)
             => (ix,ix)
             -> [(ix,e)]
             -> IO (IOToDiffArray a ix e)
newDiffArray ixs ies = do
    a <- newArray_ ixs
    sequence_ [writeArray a i e | (i, e) <- ies]
    var <- newMVar (Current a)
    return (DiffArray var)

indexDiffArray :: (MArray a e IO, Ix ix)
               => IOToDiffArray a ix e
               -> ix
               -> IO e
a `indexDiffArray` i = do
    d <- readMVar (varDiffArray a)
    case d of
        Current a'  -> readArray a' i
        Diff a' ies -> maybe (indexDiffArray a' i) return (lookup i ies)

updateDiffArray :: (MArray a e IO, Ix ix)
                => IOToDiffArray a ix e
                -> [(ix,e)]
                -> IO (IOToDiffArray a ix e)
a `updateDiffArray` ies = do
    d <- takeMVar (varDiffArray a)
    case d of
        Current a' -> case ies of
            [] -> do
                -- We don't do the copy when there is nothing to change
                -- and this is the current version. But see below.
                putMVar (varDiffArray a) d
                return a
            _:_ -> do
                diff <- sequence [do e <- readArray a' i; return (i, e)
                    | (i, _) <- ies]
                sequence_ [writeArray a' i e | (i, e) <- ies]
                var' <- newMVar (Current a')
                putMVar (varDiffArray a) (Diff (DiffArray var') diff)
                return (DiffArray var')
        Diff _ _ -> do
            -- We still do the copy when there is nothing to change
            -- but this is not the current version. So you can use
            -- 'a // []' to make sure that the resulting array has
            -- fast element access.
            putMVar (varDiffArray a) d
            a' <- thawDiffArray a
                -- thawDiffArray gives a fresh array which we can
                -- safely mutate.
            sequence_ [writeArray a' i e | (i, e) <- ies]
            var' <- newMVar (Current a')
            return (DiffArray var')

boundsDiffArray :: (HasBounds a, Ix ix)
                => IOToDiffArray a ix e
                -> IO (ix,ix)
boundsDiffArray a = do
    d <- readMVar (varDiffArray a)
    case d of
        Current a' -> return (bounds a')
        Diff a' _  -> boundsDiffArray a'

freezeDiffArray :: (MArray a e IO, Ix ix)
                => a ix e
                -> IO (IOToDiffArray a ix e)
freezeDiffArray a = do
    a' <- newArray_ (bounds a)
    sequence_ [readArray a i >>= writeArray a' i | i <- indices a]
    var <- newMVar (Current a')
    return (DiffArray var)

{-# RULES
"freeze/DiffArray"   freeze = freezeDiffArray
    #-}

-- unsafeFreezeDiffArray is really unsafe. Better don't use the old
-- array at all after freezing. The contents of the source array will
-- be changed when '//' is applied to the resulting array.

unsafeFreezeDiffArray :: (MArray a e IO, Ix ix)
                      => a ix e
                      -> IO (IOToDiffArray a ix e)
unsafeFreezeDiffArray a = do
    var <- newMVar (Current a)
    return (DiffArray var)

{-# RULES
"unsafeFreeze/DiffArray"   unsafeFreeze = unsafeFreezeDiffArray
    #-}

thawDiffArray :: (MArray a e IO, Ix ix)
              => IOToDiffArray a ix e
              -> IO (a ix e)
thawDiffArray a = do
    d <- readMVar (varDiffArray a)
    case d of
        Current a' -> do
            a'' <- newArray_ (bounds a')
            sequence_ [readArray a' i >>= writeArray a'' i | i <- indices a']
            return a''
        Diff a' ies -> do
            a'' <- thawDiffArray a'
            sequence_ [writeArray a'' i e | (i, e) <- ies]
            return a''

{-# RULES
"thaw/DiffArray"   thaw = thawDiffArray
    #-}

-- unsafeThawDiffArray is really unsafe. Better don't use the old
-- array at all after thawing. The contents of the resulting array
-- will be changed when '//' is applied to the source array.

unsafeThawDiffArray :: (MArray a e IO, Ix ix)
                    => IOToDiffArray a ix e
                    -> IO (a ix e)
unsafeThawDiffArray a = do
    d <- readMVar (varDiffArray a)
    case d of
        Current a'  -> return a'
        Diff a' ies -> do
            a'' <- unsafeThawDiffArray a'
            sequence_ [writeArray a'' i e | (i, e) <- ies]
            return a''

{-# RULES
"unsafeThaw/DiffArray"   unsafeThaw = unsafeThawDiffArray
    #-}

------------------------------------------------------------------------
-- readMVar copied from module MVar which is not available here.

readMVar :: MVar a -> IO a
readMVar var = block $ do
    a <- takeMVar var
    putMVar var a
    return a
